

use strict;
use warnings;
use 5.008001;

{
    package DBD::Pg;

    use version; our $VERSION = qv('3.16.3');

    use DBI 1.614 ();
    use Exporter ();
    use XSLoader;
    our $dbh;
    our @ISA = qw(Exporter);

    use constant {
        PG_MIN_SMALLINT => -32768,
        PG_MAX_SMALLINT => 32767,
        PG_MIN_INTEGER  => -2147483648,
        PG_MAX_INTEGER  => 2147483647,
        PG_MIN_BIGINT => '-9223372036854775808',
        PG_MAX_BIGINT => '9223372036854775807',
        PG_MIN_SMALLSERIAL => 1,
        PG_MAX_SMALLSERIAL => 32767,
        PG_MIN_SERIAL => 1,
        PG_MAX_SERIAL => 2147483647,
        PG_MIN_BIGSERIAL => 1,
        PG_MAX_BIGSERIAL => '9223372036854775807',
    };

    our %EXPORT_TAGS =
        (
         async => [qw($DBDPG_DEFAULT PG_ASYNC PG_OLDQUERY_CANCEL PG_OLDQUERY_WAIT)],
         pg_limits => [qw($DBDPG_DEFAULT
                       PG_MIN_SMALLINT PG_MAX_SMALLINT PG_MIN_INTEGER PG_MAX_INTEGER PG_MAX_BIGINT PG_MIN_BIGINT
                       PG_MIN_SMALLSERIAL PG_MAX_SMALLSERIAL PG_MIN_SERIAL PG_MAX_SERIAL PG_MIN_BIGSERIAL PG_MAX_BIGSERIAL)],
         pg_types => [qw($DBDPG_DEFAULT PG_ASYNC PG_OLDQUERY_CANCEL PG_OLDQUERY_WAIT
            PG_ACLITEM PG_ACLITEMARRAY PG_ANY PG_ANYARRAY PG_ANYCOMPATIBLE
            PG_ANYCOMPATIBLEARRAY PG_ANYCOMPATIBLEMULTIRANGE PG_ANYCOMPATIBLENONARRAY PG_ANYCOMPATIBLERANGE PG_ANYELEMENT
            PG_ANYENUM PG_ANYMULTIRANGE PG_ANYNONARRAY PG_ANYRANGE PG_BIT
            PG_BITARRAY PG_BOOL PG_BOOLARRAY PG_BOX PG_BOXARRAY
            PG_BPCHAR PG_BPCHARARRAY PG_BYTEA PG_BYTEAARRAY PG_CHAR
            PG_CHARARRAY PG_CID PG_CIDARRAY PG_CIDR PG_CIDRARRAY
            PG_CIRCLE PG_CIRCLEARRAY PG_CSTRING PG_CSTRINGARRAY PG_DATE
            PG_DATEARRAY PG_DATEMULTIRANGE PG_DATEMULTIRANGEARRAY PG_DATERANGE PG_DATERANGEARRAY
            PG_EVENT_TRIGGER PG_FDW_HANDLER PG_FLOAT4 PG_FLOAT4ARRAY PG_FLOAT8
            PG_FLOAT8ARRAY PG_GTSVECTOR PG_GTSVECTORARRAY PG_INDEX_AM_HANDLER PG_INET
            PG_INETARRAY PG_INT2 PG_INT2ARRAY PG_INT2VECTOR PG_INT2VECTORARRAY
            PG_INT4 PG_INT4ARRAY PG_INT4MULTIRANGE PG_INT4MULTIRANGEARRAY PG_INT4RANGE
            PG_INT4RANGEARRAY PG_INT8 PG_INT8ARRAY PG_INT8MULTIRANGE PG_INT8MULTIRANGEARRAY
            PG_INT8RANGE PG_INT8RANGEARRAY PG_INTERNAL PG_INTERVAL PG_INTERVALARRAY
            PG_JSON PG_JSONARRAY PG_JSONB PG_JSONBARRAY PG_JSONPATH
            PG_JSONPATHARRAY PG_LANGUAGE_HANDLER PG_LINE PG_LINEARRAY PG_LSEG
            PG_LSEGARRAY PG_MACADDR PG_MACADDR8 PG_MACADDR8ARRAY PG_MACADDRARRAY
            PG_MONEY PG_MONEYARRAY PG_NAME PG_NAMEARRAY PG_NUMERIC
            PG_NUMERICARRAY PG_NUMMULTIRANGE PG_NUMMULTIRANGEARRAY PG_NUMRANGE PG_NUMRANGEARRAY
            PG_OID PG_OIDARRAY PG_OIDVECTOR PG_OIDVECTORARRAY PG_PATH
            PG_PATHARRAY PG_PG_ATTRIBUTE PG_PG_ATTRIBUTEARRAY PG_PG_BRIN_BLOOM_SUMMARY PG_PG_BRIN_MINMAX_MULTI_SUMMARY
            PG_PG_CLASS PG_PG_CLASSARRAY PG_PG_DDL_COMMAND PG_PG_DEPENDENCIES PG_PG_LSN
            PG_PG_LSNARRAY PG_PG_MCV_LIST PG_PG_NDISTINCT PG_PG_NODE_TREE PG_PG_PROC
            PG_PG_PROCARRAY PG_PG_SNAPSHOT PG_PG_SNAPSHOTARRAY PG_PG_TYPE PG_PG_TYPEARRAY
            PG_POINT PG_POINTARRAY PG_POLYGON PG_POLYGONARRAY PG_RECORD
            PG_RECORDARRAY PG_REFCURSOR PG_REFCURSORARRAY PG_REGCLASS PG_REGCLASSARRAY
            PG_REGCOLLATION PG_REGCOLLATIONARRAY PG_REGCONFIG PG_REGCONFIGARRAY PG_REGDICTIONARY
            PG_REGDICTIONARYARRAY PG_REGNAMESPACE PG_REGNAMESPACEARRAY PG_REGOPER PG_REGOPERARRAY
            PG_REGOPERATOR PG_REGOPERATORARRAY PG_REGPROC PG_REGPROCARRAY PG_REGPROCEDURE
            PG_REGPROCEDUREARRAY PG_REGROLE PG_REGROLEARRAY PG_REGTYPE PG_REGTYPEARRAY
            PG_TABLE_AM_HANDLER PG_TEXT PG_TEXTARRAY PG_TID PG_TIDARRAY
            PG_TIME PG_TIMEARRAY PG_TIMESTAMP PG_TIMESTAMPARRAY PG_TIMESTAMPTZ
            PG_TIMESTAMPTZARRAY PG_TIMETZ PG_TIMETZARRAY PG_TRIGGER PG_TSMULTIRANGE
            PG_TSMULTIRANGEARRAY PG_TSM_HANDLER PG_TSQUERY PG_TSQUERYARRAY PG_TSRANGE
            PG_TSRANGEARRAY PG_TSTZMULTIRANGE PG_TSTZMULTIRANGEARRAY PG_TSTZRANGE PG_TSTZRANGEARRAY
            PG_TSVECTOR PG_TSVECTORARRAY PG_TXID_SNAPSHOT PG_TXID_SNAPSHOTARRAY PG_UNKNOWN
            PG_UUID PG_UUIDARRAY PG_VARBIT PG_VARBITARRAY PG_VARCHAR
            PG_VARCHARARRAY PG_VOID PG_XID PG_XID8 PG_XID8ARRAY
            PG_XIDARRAY PG_XML PG_XMLARRAY
        )],
    );

    {
        package DBD::Pg::DefaultValue;
        sub new { my $self = {}; return bless $self, shift; }
    }
    our $DBDPG_DEFAULT = DBD::Pg::DefaultValue->new();
    Exporter::export_ok_tags('pg_types', 'async', 'pg_limits');
    our @EXPORT = qw($DBDPG_DEFAULT PG_ASYNC PG_OLDQUERY_CANCEL PG_OLDQUERY_WAIT PG_BYTEA);
    XSLoader::load(__PACKAGE__, $VERSION);

    our $err = 0;       # holds error code for DBI::err
    our $errstr = '';   # holds error string for DBI::errstr
    our $sqlstate = ''; # holds five character SQLSTATE code
    our $drh = undef;   # holds driver handle once initialized

    ## These two methods are here to allow calling before connect()
    sub parse_trace_flag {
        my ($class, $flag) = @_;
        return (0x7FFFFF00 - 0x08000000) if $flag eq 'DBD'; ## all but the prefix
        return 0x01000000 if $flag eq 'pglibpq';
        return 0x02000000 if $flag eq 'pgstart';
        return 0x04000000 if $flag eq 'pgend';
        return 0x08000000 if $flag eq 'pgprefix';
        return 0x10000000 if $flag eq 'pglogin';
        return 0x20000000 if $flag eq 'pgquote';
        return DBI::parse_trace_flag($class, $flag);
    }
    sub parse_trace_flags {
        my ($class, $flags) = @_;
        return DBI::parse_trace_flags($class, $flags);
    }

    ## Both CLONE and driver are required by DBI, see perldoc DBI::DBD

    sub CLONE {
        $drh = undef;
        return;
    }

    my $methods_are_installed = 0;

    sub driver {

        return $drh if defined $drh;

        my $class = shift;

        $class .= '::dr';

        ## Work around for issue found in https://rt.cpan.org/Ticket/Display.html?id=83057
        my $realversion = qv('3.16.3');

        $drh = DBI::_new_drh($class, {
            'Name'        => 'Pg',
            'Version'     => $realversion,
            'Err'         => \$DBD::Pg::err,
            'Errstr'      => \$DBD::Pg::errstr,
            'State'       => \$DBD::Pg::sqlstate,
            'Attribution' => "DBD::Pg $realversion by Greg Sabino Mullane and others",
        });

        # uncoverable branch false
        if (!$methods_are_installed) {
            DBD::Pg::db->install_method('pg_cancel');
            DBD::Pg::db->install_method('pg_endcopy');
            DBD::Pg::db->install_method('pg_error_field');
            DBD::Pg::db->install_method('pg_getline');
            DBD::Pg::db->install_method('pg_getcopydata');
            DBD::Pg::db->install_method('pg_getcopydata_async');
            DBD::Pg::db->install_method('pg_notifies');
            DBD::Pg::db->install_method('pg_putcopydata');
            DBD::Pg::db->install_method('pg_putcopyend');
            DBD::Pg::db->install_method('pg_ping');
            DBD::Pg::db->install_method('pg_putline');
            DBD::Pg::db->install_method('pg_ready');
            DBD::Pg::db->install_method('pg_release');
            DBD::Pg::db->install_method('pg_result'); ## NOT duplicated below!
            DBD::Pg::db->install_method('pg_rollback_to');
            DBD::Pg::db->install_method('pg_savepoint');
            DBD::Pg::db->install_method('pg_server_trace');
            DBD::Pg::db->install_method('pg_server_untrace');
            DBD::Pg::db->install_method('pg_type_info');

            DBD::Pg::st->install_method('pg_cancel');
            DBD::Pg::st->install_method('pg_result');
            DBD::Pg::st->install_method('pg_ready');
            DBD::Pg::st->install_method('pg_canonical_ids');
            DBD::Pg::st->install_method('pg_canonical_names');

            DBD::Pg::db->install_method('pg_lo_creat');
            DBD::Pg::db->install_method('pg_lo_open');
            DBD::Pg::db->install_method('pg_lo_write');
            DBD::Pg::db->install_method('pg_lo_read');
            DBD::Pg::db->install_method('pg_lo_lseek');
            DBD::Pg::db->install_method('pg_lo_lseek64');
            DBD::Pg::db->install_method('pg_lo_tell');
            DBD::Pg::db->install_method('pg_lo_tell64');
            DBD::Pg::db->install_method('pg_lo_truncate');
            DBD::Pg::db->install_method('pg_lo_truncate64');
            DBD::Pg::db->install_method('pg_lo_close');
            DBD::Pg::db->install_method('pg_lo_unlink');
            DBD::Pg::db->install_method('pg_lo_import');
            DBD::Pg::db->install_method('pg_lo_import_with_oid');
            DBD::Pg::db->install_method('pg_lo_export');

            $methods_are_installed++;
        }

        return $drh;

    } ## end of driver


    1;

} ## end of package DBD::Pg


{
    package DBD::Pg::dr;

    use strict;

    ## Returns an array of formatted database names from the pg_database table
    sub data_sources {

        my $drh = shift;
        my $conninfo = shift || '';
        my $connstring = 'dbname=postgres';
        if ($ENV{DBI_DSN}) {
            ($connstring = $ENV{DBI_DSN}) =~ s/dbi:Pg://i;
        }
        if (length $conninfo) {
            $connstring .= ";$conninfo";
        }

        my $dbh = DBD::Pg::dr::connect($drh, $connstring) or return;
        $dbh->{AutoCommit}=1;
        my $SQL = 'SELECT pg_catalog.quote_ident(datname) FROM pg_catalog.pg_database ORDER BY 1';
        my $sth = $dbh->prepare($SQL);
        $sth->execute();
        $conninfo and $conninfo = ";$conninfo";
        my @sources = map { "dbi:Pg:dbname=$_->[0]$conninfo" } @{$sth->fetchall_arrayref()};
        $dbh->disconnect;
        return @sources;
    }


    sub connect { ## no critic (ProhibitBuiltinHomonyms)

        my ($drh, $dbname, $user, $pass, $attr) = @_;

        ## Allow "db" and "database" as synonyms for "dbname"
        $dbname =~ s/\b(?:db|database)\s*=/dbname=/;

        my $name = $dbname;
        if ($dbname =~ m{dbname\s*=\s*[\"\']([^\"\']+)}) {
            $name = "'$1'";
            $dbname =~ s/\"/\'/g;
        }
        elsif ($dbname =~ m{dbname\s*=\s*([^;]+)}) {
            $name = $1;
        }

        $user = defined($user) ? $user : defined $ENV{DBI_USER} ? $ENV{DBI_USER} : '';
        $pass = defined($pass) ? $pass : defined $ENV{DBI_PASS} ? $ENV{DBI_PASS} : '';

        my ($dbh) = DBI::_new_dbh($drh, {
            'Name'         => $dbname,
            'Username'     => $user,
            'CURRENT_USER' => $user,
         });

        # Connect to the database..
        DBD::Pg::db::_login($dbh, $dbname, $user, $pass, $attr) or return undef;

        my $version = $dbh->{pg_server_version};
        $dbh->{private_dbdpg}{version} = $version;

        if ($attr) {
            if ($attr->{dbd_verbose}) {
                $dbh->trace('DBD');
            }
        }

        return $dbh;
    }

    sub private_attribute_info {
        return {
        };
    }

} ## end of package DBD::Pg::dr


{
    package DBD::Pg::db;

    use DBI qw(:sql_types);

    use strict;

    sub parse_trace_flag {
        return DBD::Pg->parse_trace_flag($_[1]);
    }

    sub prepare {
        my($dbh, $statement, @attribs) = @_;

        return undef if ! defined $statement;

        # Create a 'blank' statement handle:
        my $sth = DBI::_new_sth($dbh, {
            'Statement' => $statement,
        });

        DBD::Pg::st::_prepare($sth, $statement, @attribs);

        return $sth;
    }

    sub last_insert_id {

        my ($dbh, undef, $schema, $table, undef, $attr) = @_;

        ## Our ultimate goal is to get a sequence
        my ($sth, $count, $SQL, $sequence);

        ## Cache all of our table lookups? Default is yes
        my $cache = 1;

        ## Catalog and col (arguments 2 and 5) are not used
        $schema = '' if ! defined $schema;
        $table = '' if ! defined $table;
        my $cachename = join("\0", 'lii', $schema, $table);

        if (defined $attr and length $attr) {
            ## If not a hash, assume it is a sequence name
            if (! ref $attr) {
                $attr = {sequence => $attr};
            }
            elsif (ref $attr ne 'HASH') {
                $dbh->set_err(1, 'last_insert_id must be passed a hashref as the final argument');
                return undef;
            }
            ## Named sequence overrides any table or schema settings
            if (exists $attr->{sequence} and length $attr->{sequence}) {
                $sequence = $attr->{sequence};
            }
            if (exists $attr->{pg_cache}) {
                $cache = $attr->{pg_cache};
            }
        }

        if (! defined $sequence and exists $dbh->{private_dbdpg}{$cachename} and $cache) {
            $sequence = $dbh->{private_dbdpg}{$cachename};
        }
        elsif (! defined $sequence) {
            ## At this point, we must have a valid table name
            if (! length $table) {
                $dbh->set_err(1, 'last_insert_id needs at least a sequence or table name');
                return undef;
            }
            my @args = ($table);
            my $schemawhere;
            if (length $schema) {
                # if given a schema, use that
                $schemawhere = 'n.nspname = ?';
                push @args, $schema;
            } else {
                # otherwise it must be visible via the search path
                $schemawhere = 'pg_catalog.pg_table_is_visible(c.oid)';
            }
            ## Is there a sequence associated with the table via a unique, indexed column,
            ## either via ownership (e.g. serial, identity) or a manual default?
            my $idcond = $dbh->{private_dbdpg}{version} >= 100000
                ? q{a.attidentity <> ''} : q{false};
            $SQL = sprintf(q{
                SELECT i.indisprimary,
                    COALESCE(
                        -- this takes the table name as text, not regclass
                        pg_catalog.pg_get_serial_sequence(
                            -- and pre-8.3 doesn't have a cast from regclass to text,
                            -- and pre-9.3 doesn't have format, so do it the long way
                            quote_ident(n.nspname) || '.' || quote_ident(c.relname),
                            a.attname),
                        (SELECT replace(substring(pg_catalog.pg_get_expr(d.adbin, d.adrelid)
                                            from $r$^nextval\('(.+)'::[\w\s]+\)$$r$),
                                        -- unescape any single quotes from the default
                                        $$''$$, $$'$$)
                            FROM pg_catalog.pg_attrdef d
                            WHERE a.atthasdef
                                AND a.attrelid = d.adrelid
                                AND a.attnum = d.adnum)
                    ) AS seqname
                FROM pg_class c
                    JOIN pg_catalog.pg_namespace n ON (n.oid = c.relnamespace)
                    -- LEFT JOIN so we can distingiuish between table not found (zero rows)
                    -- and no suitable column found (at least one all-NULL row)
                    LEFT JOIN pg_catalog.pg_index i
                        ON c.oid = i.indrelid AND i.indisunique
                    LEFT JOIN pg_catalog.pg_attribute a
                        ON i.indrelid = a.attrelid AND i.indkey[0]=a.attnum
                        AND (a.atthasdef OR %s)
                WHERE c.relname = ? AND %s
            }, $idcond, $schemawhere);
            $sth = $dbh->prepare_cached($SQL);
            $count = $sth->execute(@args);
            if (!defined $count or $count eq '0E0') {
                $sth->finish();
                my $message = qq{Could not find the table "$table"};
                length $schema and $message .= qq{ in the schema "$schema"};
                $dbh->set_err(1, $message);
                return undef;
            }
            my $info = $sth->fetchall_arrayref();
            ## We have at least one with a default value. See if we found any sequences
            my @def = grep { defined $_->[1] } @$info;
            if (!@def) {
                ## This may be an inherited table, in which case we can use the parent's info
                $SQL = 'SELECT inhparent::regclass FROM pg_inherits WHERE inhrelid = ?::regclass::oid';
                my $isth = $dbh->prepare($SQL);
                $count = $isth->execute($table);
                if ($count < 1) {
                    $isth->finish();
                    $dbh->set_err(1, qq{No suitable column found for last_insert_id of table "$table"\n});
                    return undef;
                }
                my $parent = $isth->fetch->[0];
                $args[0] = $parent;
                $count = $sth->execute(@args);
                if (1 == $count) {
                    $info = $sth->fetchall_arrayref();
                    @def = grep { defined $_->[1] } @$info;
                }
                if (!@def) {
                    $sth->finish();
                    $dbh->set_err(1, qq{No suitable column found for last_insert_id of table "$table"\n});
                    return undef;
                }
                ## Fall through with inherited information
            }
            ## Tiebreaker goes to the primary keys
            if (@def > 1) {
                my @pri = grep { $_->[0] } @def;
                if (1 != @pri) {
                    $dbh->set_err(1, qq{No suitable column found for last_insert_id of table "$table"\n});
                    return undef;
                }
                @def = @pri;
            }
            $sequence = $def[0]->[1];
            ## Cache this information for subsequent calls
            $dbh->{private_dbdpg}{$cachename} = $sequence;
        }

        $sth = $dbh->prepare_cached('SELECT pg_catalog.currval(?)');
        $count = $sth->execute($sequence);
        return undef if ! defined $count;
        return $sth->fetchall_arrayref()->[0][0];

    } ## end of last_insert_id

    sub ping {
        my $dbh = shift;
        local $SIG{__WARN__} if $dbh->FETCH('PrintError');
        my $ret = DBD::Pg::db::_ping($dbh);
        return $ret < 1 ? 0 : $ret;
    }

    sub pg_ping {
        my $dbh = shift;
        local $SIG{__WARN__} if $dbh->FETCH('PrintError');
        return DBD::Pg::db::_ping($dbh);
    }

    sub pg_type_info {
        my($dbh,$pg_type) = @_;
        local $SIG{__WARN__} if $dbh->FETCH('PrintError');
        return DBD::Pg::db::_pg_type_info($pg_type);
    }

    # Column expected in statement handle returned.
    # table_cat, table_schem, table_name, column_name, data_type, type_name,
     # column_size, buffer_length, DECIMAL_DIGITS, NUM_PREC_RADIX, NULLABLE,
    # REMARKS, COLUMN_DEF, SQL_DATA_TYPE, SQL_DATETIME_SUB, CHAR_OCTET_LENGTH,
    # ORDINAL_POSITION, IS_NULLABLE
    # The result set is ordered by TABLE_SCHEM, TABLE_NAME and ORDINAL_POSITION.

    sub column_info {
        my $dbh = shift;
        my (undef, $schema, $table, $column) = @_;

        my @search;
        ## If the schema or table has an underscore or a %, use a LIKE comparison
        if (defined $schema and length $schema) {
            push @search, 'n.nspname ' . ($schema =~ /[_%]/ ? 'LIKE ' : '= ') .
                $dbh->quote($schema);
        }
        if (defined $table and length $table) {
            push @search, 'c.relname ' . ($table =~ /[_%]/ ? 'LIKE ' : '= ') .
                $dbh->quote($table);
        }
        if (defined $column and length $column) {
            push @search, 'a.attname ' . ($column =~ /[_%]/ ? 'LIKE ' : '= ') .
                $dbh->quote($column);
        }

        my $whereclause = join "\n\t\t\t\tAND ", '', @search;

        my $col_info_sql = qq!
            SELECT
                pg_catalog.quote_ident(pg_catalog.current_database()) AS "TABLE_CAT"
                , pg_catalog.quote_ident(n.nspname) AS "TABLE_SCHEM"
                , pg_catalog.quote_ident(c.relname) AS "TABLE_NAME"
                , pg_catalog.quote_ident(a.attname) AS "COLUMN_NAME"
                , a.atttypid AS "DATA_TYPE"
                , pg_catalog.format_type(a.atttypid, NULL) AS "TYPE_NAME"
                , a.attlen AS "COLUMN_SIZE"
                , NULL::text AS "BUFFER_LENGTH"
                , NULL::text AS "DECIMAL_DIGITS"
                , NULL::text AS "NUM_PREC_RADIX"
                , CASE a.attnotnull WHEN 't' THEN 0 ELSE 1 END AS "NULLABLE"
                , pg_catalog.col_description(a.attrelid, a.attnum) AS "REMARKS"
                , pg_catalog.pg_get_expr(af.adbin, af.adrelid) AS "COLUMN_DEF"
                , NULL::text AS "SQL_DATA_TYPE"
                , NULL::text AS "SQL_DATETIME_SUB"
                , NULL::text AS "CHAR_OCTET_LENGTH"
                , a.attnum AS "ORDINAL_POSITION"
                , CASE a.attnotnull WHEN 't' THEN 'NO' ELSE 'YES' END AS "IS_NULLABLE"
                , pg_catalog.format_type(a.atttypid, a.atttypmod) AS "pg_type"
                , '?' AS "pg_constraint"
                , n.nspname AS "pg_schema"
                , c.relname AS "pg_table"
                , a.attname AS "pg_column"
                , a.attrelid AS "pg_attrelid"
                , a.attnum AS "pg_attnum"
                , a.atttypmod AS "pg_atttypmod"
                , t.typtype AS "_pg_type_typtype"
                , t.oid AS "_pg_type_oid"
            FROM
                pg_catalog.pg_type t
                JOIN pg_catalog.pg_attribute a ON (t.oid = a.atttypid)
                JOIN pg_catalog.pg_class c ON (a.attrelid = c.oid)
                LEFT JOIN pg_catalog.pg_attrdef af ON (a.attnum = af.adnum AND a.attrelid = af.adrelid)
                JOIN pg_catalog.pg_namespace n ON (n.oid = c.relnamespace)
            WHERE
                a.attnum >= 0
                AND c.relkind IN ('r','p','v','m','f')
                $whereclause
            ORDER BY "TABLE_SCHEM", "TABLE_NAME", "ORDINAL_POSITION"
            !;

        my $data = $dbh->selectall_arrayref($col_info_sql);

        # To turn the data back into a statement handle, we need 
        # to fetch the data as an array of arrays, and also have a
        # a matching array of all the column names
        my %col_map = (qw/
            TABLE_CAT             0
            TABLE_SCHEM           1
            TABLE_NAME            2
            COLUMN_NAME           3
            DATA_TYPE             4
            TYPE_NAME             5
            COLUMN_SIZE           6
            BUFFER_LENGTH         7
            DECIMAL_DIGITS        8
            NUM_PREC_RADIX        9
            NULLABLE             10
            REMARKS              11
            COLUMN_DEF           12
            SQL_DATA_TYPE        13
            SQL_DATETIME_SUB     14
            CHAR_OCTET_LENGTH    15
            ORDINAL_POSITION     16
            IS_NULLABLE          17
            pg_type              18
            pg_constraint        19
            pg_schema            20
            pg_table             21
            pg_column            22
            pg_enum_values       23
            /);

        for my $row (@$data) {
            my $typoid = pop @$row;
            my $typtype = pop @$row;
            my $typmod = pop @$row;
            my $attnum = pop @$row;
            my $aid = pop @$row;

            $row->[$col_map{COLUMN_SIZE}] =
                 _calc_col_size($typmod,$row->[$col_map{COLUMN_SIZE}]);

            # Replace the Pg type with the SQL_ type
            $row->[$col_map{DATA_TYPE}] = DBD::Pg::db::pg_type_info($dbh,$row->[$col_map{DATA_TYPE}]);

            # Add pg_constraint
            my $SQL = q{SELECT pg_catalog.pg_get_constraintdef(oid) }.
                q{FROM pg_catalog.pg_constraint WHERE contype = 'c' AND }.
                qq{conrelid = $aid AND conkey = '{$attnum}'};
            my $info = $dbh->selectall_arrayref($SQL);
            if (@$info) {
                $row->[$col_map{pg_constraint}] = $info->[0][0];
            }
            else {
                $row->[$col_map{pg_constraint}] = undef;
            }

            if ( $typtype eq 'e' ) {
                my $order_column = $dbh->{private_dbdpg}{version} >= 90100
                    ? 'enumsortorder' : 'oid';
                $SQL = "SELECT enumlabel FROM pg_catalog.pg_enum WHERE enumtypid = $typoid ORDER BY $order_column";
                $row->[$col_map{pg_enum_values}] = $dbh->selectcol_arrayref($SQL);
            }
            else {
                $row->[$col_map{pg_enum_values}] = undef;
            }
        }

        # Since we've processed the data in Perl, we have to jump through a hoop
        # To turn it back into a statement handle
        #
        return _prepare_from_data(
             'column_info',
             $data,
             [ sort { $col_map{$a} <=> $col_map{$b} } keys %col_map],
        );
    }

    sub _prepare_from_data {
        my ($statement, $data, $names, %attrinfo) = @_;
        my $sponge = DBI->connect('dbi:Sponge:', '', '', { RaiseError => 1 });
        my $sth = $sponge->prepare($statement, { rows=>$data, NAME=>$names, %attrinfo });
        return $sth;
    }

    sub statistics_info {

        my $dbh = shift;
        my (undef, $schema, $table, $unique_only) = @_;

        ## Catalog is ignored, but table is mandatory
        return undef unless defined $table and length $table;

        my $schema_where = '';
        my @exe_args = ($table);

        my $input_schema = (defined $schema and length $schema) ? 1 : 0;

        if ($input_schema) {
            $schema_where = 'AND n.nspname = ?';
            push(@exe_args, $schema);
        }

        my $stats_sql;

        # Table-level stats
        if (!$unique_only) {
            $stats_sql .= qq{
                SELECT
                    pg_catalog.current_database() AS "TABLE_CAT",
                    n.nspname                     AS "TABLE_SCHEM",
                    d.relname                     AS "TABLE_NAME",
                    NULL                          AS "NON_UNIQUE",
                    NULL                          AS "INDEX_QUALIFIER",
                    NULL                          AS "INDEX_NAME",
                    'table'                       AS "TYPE",
                    NULL                          AS "ORDINAL_POSITION",
                    NULL                          AS "COLUMN_NAME",
                    NULL                          AS "ASC_OR_DESC",
                    d.reltuples                   AS "CARDINALITY",
                    d.relpages                    AS "PAGES",
                    NULL                          AS "FILTER_CONDITION",
                    NULL                          AS "pg_expression",
                    NULL                          AS "pg_is_key_column",
                    NULL                          AS "pg_null_ordering"
                FROM   pg_catalog.pg_class d
                JOIN   pg_catalog.pg_namespace n ON n.oid = d.relnamespace
                WHERE  d.relname = ? $schema_where
                UNION ALL
            };
            push @exe_args, @exe_args;
        }

        my $is_key_column = $dbh->{private_dbdpg}{version} >= 110000
            ? 'col.i <= i.indnkeyatts' : 'true';

        my ($asc_or_desc, $null_ordering);
        if ($dbh->{private_dbdpg}{version} >= 90600) {
            $asc_or_desc = q{
                CASE WHEN pg_catalog.pg_index_column_has_property(c.oid, col.i, 'asc') THEN 'A'
                     WHEN pg_catalog.pg_index_column_has_property(c.oid, col.i, 'desc') THEN 'D'
                END};
            $null_ordering = q{
                CASE WHEN pg_catalog.pg_index_column_has_property(c.oid, col.i, 'nulls_first') THEN 'first'
                     WHEN pg_catalog.pg_index_column_has_property(c.oid, col.i, 'nulls_last') THEN 'last'
                END};
        }
        elsif ($dbh->{private_dbdpg}{version} > 80300) {
            $asc_or_desc = q{
                CASE WHEN a.amcanorder THEN
                     CASE WHEN i.indoption[col.i - 1] & 1 = 0 THEN 'A' ELSE 'D' END
                END};
            $null_ordering = q{
                CASE WHEN a.amcanorder THEN
                     CASE WHEN i.indoption[col.i - 1] & 2 = 0 THEN 'last' ELSE 'first' END
                END};
        }
        else {
            $asc_or_desc = q{CASE WHEN a.amorderstrategy <> 0 THEN 'A' END};
            $null_ordering = q{CASE WHEN a.amorderstrategy <> 0 THEN 'last' END};
        }

        # Fetch the index definitions
        $stats_sql .= qq{
            SELECT
                pg_catalog.current_database() AS "TABLE_CAT",
                n.nspname                     AS "TABLE_SCHEM",
                d.relname                     AS "TABLE_NAME",
                NOT(i.indisunique)            AS "NON_UNIQUE",
                NULL                          AS "INDEX_QUALIFIER",
                c.relname                     AS "INDEX_NAME",
                CASE WHEN i.indisclustered THEN 'clustered'
                     WHEN a.amname = 'btree' THEN 'btree'
                     WHEN a.amname = 'hash' THEN 'hashed'
                     ELSE 'other'
                END                           AS "TYPE",
                col.i                         AS "ORDINAL_POSITION",
                att.attname                   AS "COLUMN_NAME",
                $asc_or_desc                  AS "ASC_OR_DESC",
                c.reltuples                   AS "CARDINALITY",
                c.relpages                    AS "PAGES",
                pg_catalog.pg_get_expr(i.indpred,i.indrelid)
                                              AS "FILTER_CONDITION",
                pg_catalog.pg_get_indexdef(i.indexrelid, col.i, true)
                                              AS "pg_expression",
                $is_key_column                AS "pg_is_key_column",
                $null_ordering                AS "pg_null_ordering"
            FROM
                pg_catalog.pg_index i
                JOIN pg_catalog.pg_class c ON c.oid = i.indexrelid
                JOIN pg_catalog.pg_class d ON d.oid = i.indrelid
                JOIN pg_catalog.pg_am a ON a.oid = c.relam
                JOIN pg_catalog.pg_namespace n ON n.oid = d.relnamespace
                JOIN pg_catalog.generate_series(1, pg_catalog.current_setting('max_index_keys')::integer) col(i)
                     ON col.i <= i.indnatts
                LEFT JOIN pg_catalog.pg_attribute att
                     ON att.attrelid = d.oid AND att.attnum = i.indkey[col.i - 1]
            WHERE
                d.relname = ? $schema_where
                AND (i.indisunique OR NOT(?)) -- unique_only
            ORDER BY
                "NON_UNIQUE", "TYPE", "INDEX_QUALIFIER", "INDEX_NAME", "ORDINAL_POSITION"
        };

        my $sth = $dbh->prepare($stats_sql);
        $sth->execute(@exe_args, 0+!!$unique_only);
        return $sth;
    }

    sub primary_key_info {

        my $dbh = shift;
        my (undef, $schema, $table, $attr) = @_;

        my @cols = (qw(TABLE_CAT TABLE_SCHEM TABLE_NAME
                       COLUMN_NAME KEY_SEQ PK_NAME DATA_TYPE
                       pg_tablespace_name pg_tablespace_location
                       pg_schema pg_table pg_column
                       )
            );

        ## Catalog is ignored, but table is mandatory
        if (! defined $table or ! length $table) {
            return _prepare_from_data('primary_key_info', [], \@cols);
        }

        my $whereclause = 'AND c.relname = ' . $dbh->quote($table);

        if (defined $schema and length $schema) {
            $whereclause .= "\n\t\t\tAND n.nspname = " . $dbh->quote($schema);
        }

        my $pri_key_sql = qq{
            SELECT
                  c.oid
                , pg_catalog.quote_ident(n.nspname)
                , pg_catalog.quote_ident(c.relname)
                , pg_catalog.quote_ident(c2.relname)
                , i.indkey
                , pg_catalog.quote_ident(t.spcname)
                , pg_catalog.quote_ident(t.spclocation)
                , n.nspname, c.relname, c2.relname
                , pg_catalog.quote_ident(pg_catalog.current_database())
            FROM
                pg_catalog.pg_class c
                JOIN pg_catalog.pg_index i ON (i.indrelid = c.oid)
                JOIN pg_catalog.pg_class c2 ON (c2.oid = i.indexrelid)
                LEFT JOIN pg_catalog.pg_namespace n ON (n.oid = c.relnamespace)
                LEFT JOIN pg_catalog.pg_tablespace t ON (t.oid = c.reltablespace)
            WHERE
                i.indisprimary IS TRUE
            $whereclause
        };

        if ($dbh->{private_dbdpg}{version} >= 90200) {
            $pri_key_sql =~ s/t.spclocation/pg_catalog.pg_tablespace_location(t.oid)/;
        }

        my $sth = $dbh->prepare($pri_key_sql);
        $sth->execute();
        my $info = $sth->fetchall_arrayref()->[0];
        if (! defined $info) {
            return _prepare_from_data('primary_key_info', [], \@cols);
        }

        # Get the attribute information
        my $indkey = join ',', split /\s+/, $info->[4];
        my $sql = qq{
            SELECT a.attnum, pg_catalog.quote_ident(a.attname) AS colname,
                pg_catalog.quote_ident(t.typname) AS typename
            FROM pg_catalog.pg_attribute a, pg_catalog.pg_type t
            WHERE a.attrelid = '$info->[0]'
            AND a.atttypid = t.oid
            AND attnum IN ($indkey);
        };
        $sth = $dbh->prepare($sql);
        $sth->execute();
        my $attribinfo = $sth->fetchall_hashref('attnum');

        my $pkinfo = [];

        ## Normal way: complete "row" per column in the primary key
        if (!exists $attr->{'pg_onerow'}) {
            my $x=0;
            my @key_seq = split/\s+/, $info->[4];
            for (@key_seq) {
                # TABLE_CAT
                $pkinfo->[$x][0] = $info->[10];
                # SCHEMA_NAME
                $pkinfo->[$x][1] = $info->[1];
                # TABLE_NAME
                $pkinfo->[$x][2] = $info->[2];
                # COLUMN_NAME
                $pkinfo->[$x][3] = $attribinfo->{$_}{colname};
                # KEY_SEQ
                $pkinfo->[$x][4] = $_;
                # PK_NAME
                $pkinfo->[$x][5] = $info->[3];
                # DATA_TYPE
                $pkinfo->[$x][6] = $attribinfo->{$_}{typename};
                $pkinfo->[$x][7] = $info->[5];
                $pkinfo->[$x][8] = $info->[6];
                $pkinfo->[$x][9] = $info->[7];
                $pkinfo->[$x][10] = $info->[8];
                $pkinfo->[$x][11] = $info->[9];
                $x++;
            }
        }
        else { ## Nicer way: return only one row

            # TABLE_CAT
            $info->[0] = $info->[10];
            # TABLESPACES
            $info->[7] = $info->[5];
            $info->[8] = $info->[6];
            # Unquoted names
            $info->[9] = $info->[7];
            $info->[10] = $info->[8];
            $info->[11] = $info->[9];
            # PK_NAME
            $info->[5] = $info->[3];
            # COLUMN_NAME
            $info->[3] = 2==$attr->{'pg_onerow'} ?
                [ map { $attribinfo->{$_}{colname} } split /\s+/, $info->[4] ] :
                    join ', ', map { $attribinfo->{$_}{colname} } split /\s+/, $info->[4];
            # DATA_TYPE
            $info->[6] = 2==$attr->{'pg_onerow'} ?
                [ map { $attribinfo->{$_}{typename} } split /\s+/, $info->[4] ] :
                    join ', ', map { $attribinfo->{$_}{typename} } split /\s+/, $info->[4];
            # KEY_SEQ
            $info->[4] = 2==$attr->{'pg_onerow'} ?
                [ split /\s+/, $info->[4] ] :
                    join ', ', split /\s+/, $info->[4];

            $pkinfo = [$info];
        }

        return _prepare_from_data('primary_key_info', $pkinfo, \@cols);

    }

    sub primary_key {
        my $sth = primary_key_info(@_[0..3], {pg_onerow => 2});
        my $result = $sth->fetchall_arrayref();
        return defined $result->[0] ? @{$result->[0][3]} : ();
    }


    sub foreign_key_info {

        my $dbh = shift;

        ## PK: catalog, schema, table, FK: catalog, schema, table, attr
        ## Each of these may be undef or empty
        my $pschema = $_[1] || '';
        my $ptable = $_[2] || '';
        my $fschema = $_[4] || '';
        my $ftable = $_[5] || '';

        my @cols = (qw(
            UK_TABLE_CAT UK_TABLE_SCHEM UK_TABLE_NAME UK_COLUMN_NAME
            FK_TABLE_CAT FK_TABLE_SCHEM FK_TABLE_NAME FK_COLUMN_NAME
            ORDINAL_POSITION UPDATE_RULE DELETE_RULE FK_NAME UK_NAME
            DEFERABILITY UNIQUE_OR_PRIMARY UK_DATA_TYPE FK_DATA_TYPE
        ));

        if ($dbh->{FetchHashKeyName} eq 'NAME_lc') {
            for my $col (@cols) {
                $col = lc $col;
            }
        }

        ## Must have at least one named table
        if (!length($ptable) and !length($ftable)) {
            return _prepare_from_data('foreign_key_info', [], \@cols);
        }

        ## If only the primary table is given, we return only those columns
        ## that are used as foreign keys, even if that means that we return
        ## unique keys but not primary one. We also return all the foreign
        ## tables/columns that are referencing them, of course.
        ## If no schema is given, respect search_path by using pg_table_is_visible()
        my @where;
        for ([$ptable, $pschema, 'uk'], [$ftable, $fschema, 'fk']) {
            my ($table, $schema, $type) = @$_;
            if (length $table) {
                push @where, "${type}_class.relname = " . $dbh->quote($table);
                if (length $schema) {
                    push @where, "${type}_ns.nspname = " . $dbh->quote($schema);
                }
                else {
                    push @where, "pg_catalog.pg_table_is_visible(${type}_class.oid)"
                }
            }
        }

        my $WHERE = join ' AND ', @where;
        my $SQL = qq{
            SELECT
                pg_catalog.quote_ident(pg_catalog.current_database()),
                pg_catalog.quote_ident(uk_ns.nspname),
                pg_catalog.quote_ident(uk_class.relname),
                pg_catalog.quote_ident(uk_col.attname),
                pg_catalog.quote_ident(pg_catalog.current_database()),
                pg_catalog.quote_ident(fk_ns.nspname),
                pg_catalog.quote_ident(fk_class.relname),
                pg_catalog.quote_ident(fk_col.attname),
                colnum.i,
                CASE constr.confupdtype
                    WHEN 'c' THEN 0 WHEN 'r' THEN 1 WHEN 'n' THEN 2 WHEN 'a' THEN 3 WHEN 'd' THEN 4 ELSE -1
                END,
                CASE constr.confdeltype
                    WHEN 'c' THEN 0 WHEN 'r' THEN 1 WHEN 'n' THEN 2 WHEN 'a' THEN 3 WHEN 'd' THEN 4 ELSE -1
                END,
                pg_catalog.quote_ident(constr.conname), pg_catalog.quote_ident(uk_constr.conname),
                CASE
                    WHEN constr.condeferrable = 'f' THEN 7
                    WHEN constr.condeferred = 't' THEN 6
                    WHEN constr.condeferred = 'f' THEN 5
                    ELSE -1
                END,
                CASE coalesce(uk_constr.contype, 'u')
                    WHEN 'u' THEN 'UNIQUE' WHEN 'p' THEN 'PRIMARY'
                END,
                pg_catalog.quote_ident(uk_type.typname), pg_catalog.quote_ident(fk_type.typname)
            FROM pg_catalog.pg_constraint constr
                JOIN pg_catalog.pg_class uk_class ON constr.confrelid = uk_class.oid
                JOIN pg_catalog.pg_namespace uk_ns ON uk_class.relnamespace = uk_ns.oid
                JOIN pg_catalog.pg_class fk_class ON constr.conrelid = fk_class.oid
                JOIN pg_catalog.pg_namespace fk_ns ON fk_class.relnamespace = fk_ns.oid
                -- can't do unnest() until 8.4, and would need WITH ORDINALITY to get the array indices,
                -- wich isn't available until 9.4 at the earliest, so we join against a series table instead
                JOIN pg_catalog.generate_series(1, pg_catalog.current_setting('max_index_keys')::integer) colnum(i)
                    ON colnum.i <= pg_catalog.array_upper(constr.conkey,1)
                JOIN pg_catalog.pg_attribute uk_col ON uk_col.attrelid = constr.confrelid AND uk_col.attnum = constr.confkey[colnum.i]
                JOIN pg_catalog.pg_type uk_type ON uk_col.atttypid = uk_type.oid
                JOIN pg_catalog.pg_attribute fk_col ON fk_col.attrelid = constr.conrelid AND fk_col.attnum = constr.conkey[colnum.i]
                JOIN pg_catalog.pg_type fk_type ON fk_col.atttypid = fk_type.oid

                -- We can't match confkey from the fk constraint to conkey of the unique constraint,
                -- because the unique constraint might not exist or there might be more than one
                -- matching one. However, there must be at least a unique _index_ on the key
                -- columns, so we look for that; but we can't find it via pg_index, since there may
                -- again be more than one matching index.

                -- So instead, we look at pg_depend for the dependency that was created by the fk
                -- constraint. This dependency is of type 'n' (normal) and ties the pg_constraint
                -- row oid to the pg_class oid for the index relation (a single arbitrary one if
                -- more than one matching unique index existed at the time the constraint was
                -- created).  Fortunately, the constraint does not create dependencies on the
                -- referenced table itself, but on the _columns_ of the referenced table, so the
                -- index can be distinguished easily.  Then we look for another pg_depend entry,
                -- this time an 'i' (implementation) dependency from a pg_constraint oid (the unique
                -- constraint if one exists) to the index oid; but we have to allow for the
                -- possibility that this one doesn't exist.          - Andrew Gierth (RhodiumToad)

                JOIN pg_catalog.pg_depend dep ON (
                    dep.classid = 'pg_catalog.pg_constraint'::regclass
                    AND dep.objid = constr.oid
                    AND dep.objsubid = 0
                    AND dep.deptype = 'n'
                    AND dep.refclassid = 'pg_catalog.pg_class'::regclass
                    AND dep.refobjsubid=0
                )
                JOIN pg_catalog.pg_class idx ON (
                    idx.oid = dep.refobjid AND idx.relkind='i'
                )
                LEFT JOIN pg_catalog.pg_depend dep2 ON (
                    dep2.classid = 'pg_catalog.pg_class'::regclass
                    AND dep2.objid = idx.oid
                    AND dep2.objsubid = 0
                    AND dep2.deptype = 'i'
                    AND dep2.refclassid = 'pg_catalog.pg_constraint'::regclass
                    AND dep2.refobjsubid = 0
                )
                LEFT JOIN pg_catalog.pg_constraint uk_constr ON (
                    uk_constr.oid = dep2.refobjid AND uk_constr.contype IN ('p','u')
                )
            WHERE $WHERE
                AND uk_class.relkind ~ 'r|p'
                AND fk_class.relkind ~ 'r|p'
                AND constr.contype = 'f'
            ORDER BY constr.conname, colnum.i
        };
        my $fkinfo = $dbh->selectall_arrayref($SQL);

        return _prepare_from_data('foreign_key_info', $fkinfo, \@cols);

    } ## end of foreign_key_info


    sub table_info {

        my $dbh = shift;
        my ($catalog, $schema, $table, $type) = @_;

        my $tbl_sql = ();

        my $extracols = q{,NULL::text AS pg_schema, NULL::text AS pg_table};
        if ( # Rule 19a
                (defined $catalog and $catalog eq '%')
                and (defined $schema and $schema eq '')
                and (defined $table and $table eq '')
             ) {
            $tbl_sql = qq{
                    SELECT
                       pg_catalog.quote_ident(pg_catalog.current_database()) AS "TABLE_CAT"
                     , NULL::text AS "TABLE_SCHEM"
                     , NULL::text AS "TABLE_NAME"
                     , NULL::text AS "TABLE_TYPE"
                     , NULL::text AS "REMARKS" $extracols
                    };
        }
        elsif (# Rule 19b
                     (defined $catalog and $catalog eq '')
                     and (defined $schema and $schema eq '%')
                     and (defined $table and $table eq '')
                    ) {
            $extracols = q{,n.nspname AS pg_schema, NULL::text AS pg_table};
            $tbl_sql = qq{SELECT
                       NULL::text AS "TABLE_CAT"
                     , pg_catalog.quote_ident(n.nspname) AS "TABLE_SCHEM"
                     , NULL::text AS "TABLE_NAME"
                     , NULL::text AS "TABLE_TYPE"
                     , CASE WHEN n.nspname ~ '^pg_' THEN 'system schema' ELSE 'owned by ' || pg_catalog.pg_get_userbyid(n.nspowner) END AS "REMARKS" $extracols
                    FROM pg_catalog.pg_namespace n
                    ORDER BY "TABLE_SCHEM"
                    };
        }
        elsif (# Rule 19c
                     (defined $catalog and $catalog eq '')
                     and (defined $schema and $schema eq '')
                     and (defined $table and $table eq '')
                     and (defined $type and $type eq '%')
                    ) {
            $tbl_sql = q{
                    SELECT "TABLE_CAT"
                         , "TABLE_SCHEM"
                         , "TABLE_NAME"
                         , "TABLE_TYPE"
                         , "REMARKS"
                    FROM
                      (SELECT NULL::text AS "TABLE_CAT"
                            , NULL::text AS "TABLE_SCHEM"
                            , NULL::text AS "TABLE_NAME") dummy_cols
                    CROSS JOIN
                      (SELECT 'TABLE'        AS "TABLE_TYPE"
                            , 'relkind ~ r|p'   AS "REMARKS"
                       UNION
                       SELECT 'SYSTEM TABLE'
                            , 'relkind ~ r|p; nspname ~ ^pg_(catalog|toast)$'
                       UNION
                       SELECT 'VIEW'
                            , 'relkind: v'
                       UNION
                       SELECT 'SYSTEM VIEW'
                            , 'relkind: v; nspname ~ ^pg_(catalog|toast)$'
                       UNION
                       SELECT 'MATERIALIZED VIEW'
                            , 'relkind: m'
                       UNION
                       SELECT 'SYSTEM MATERIALIZED VIEW'
                            , 'relkind: m; nspname ~ ^pg_(catalog|toast)$'
                       UNION
                       SELECT 'FOREIGN TABLE'
                            , 'relkind: f'
                       UNION
                       SELECT 'SYSTEM FOREIGN TABLE'
                            , 'relkind: f; nspname ~ ^pg_(catalog|toast)$'
                       UNION
                       SELECT 'LOCAL TEMPORARY'
                            , 'relkind ~ r|p; nspname ~ ^pg_(toast_)?temp') type_info
                     ORDER BY "TABLE_TYPE" ASC
                };
        }
        else {
            # Default SQL
            $extracols = q{,n.nspname AS pg_schema, c.relname AS pg_table};
            my @search = (q|c.relkind IN ('r', 'p', 'v', 'm', 'f')|, # No sequences, etc. for now
                          q|NOT (pg_catalog.quote_ident(n.nspname) ~ '^pg_(toast_)?temp_' AND NOT pg_catalog.has_schema_privilege(n.nspname, 'USAGE'))|);   # No others' temp objects
            my $showtablespace = sprintf q{pg_catalog.quote_ident(%s) AS "pg_tablespace_location"},
                $dbh->{private_dbdpg}{version} < 90200 ? 't.spclocation' : 'pg_catalog.pg_tablespace_location(t.oid)';

            ## If the schema or table has an underscore or a %, use a LIKE comparison
            if (defined $schema and length $schema) {
                    push @search, 'n.nspname ' . ($schema =~ /[_%]/ ? 'LIKE ' : '= ') . $dbh->quote($schema);
            }
            if (defined $table and length $table) {
                    push @search, 'c.relname ' . ($table =~ /[_%]/ ? 'LIKE ' : '= ') . $dbh->quote($table);
            }

            my $whereclause = join "\n\t\t\t\t\t AND " => @search;
            $tbl_sql = qq{
                SELECT pg_catalog.quote_ident(pg_catalog.current_database()) AS "TABLE_CAT"
                     , pg_catalog.quote_ident(n.nspname) AS "TABLE_SCHEM"
                     , pg_catalog.quote_ident(c.relname) AS "TABLE_NAME"
                       -- any temp table or temp view is LOCAL TEMPORARY for us
                     , CASE WHEN pg_catalog.quote_ident(n.nspname) ~ '^pg_(toast_)?temp_' THEN
                                 'LOCAL TEMPORARY'
                            WHEN c.relkind ~ 'r|p' THEN
                                 CASE WHEN pg_catalog.quote_ident(n.nspname) ~ '^pg_' THEN
                                           'SYSTEM TABLE'
                                      ELSE 'TABLE'
                                  END
                            WHEN c.relkind = 'v' THEN
                                 CASE WHEN pg_catalog.quote_ident(n.nspname) ~ '^pg_' THEN
                                           'SYSTEM VIEW'
                                      ELSE 'VIEW'
                                  END
                            WHEN c.relkind = 'm' THEN
                                 CASE WHEN pg_catalog.quote_ident(n.nspname) ~ '^pg_' THEN
                                           'SYSTEM MATERIALIZED VIEW'
                                      ELSE 'MATERIALIZED VIEW'
                                  END
                            WHEN c.relkind = 'f' THEN
                                 CASE WHEN pg_catalog.quote_ident(n.nspname) ~ '^pg_' THEN
                                           'SYSTEM FOREIGN TABLE'
                                      ELSE 'FOREIGN TABLE'
                                  END
                            ELSE 'UNKNOWN'
                         END AS "TABLE_TYPE"
                     , d.description AS "REMARKS"
                     , pg_catalog.quote_ident(t.spcname) AS "pg_tablespace_name"
                     , $showtablespace $extracols
                  FROM pg_catalog.pg_class AS c
                  LEFT JOIN pg_catalog.pg_description AS d
                       ON (c.oid = d.objoid AND c.tableoid = d.classoid AND d.objsubid = 0)
                  LEFT JOIN pg_catalog.pg_namespace n ON (n.oid = c.relnamespace)
                  LEFT JOIN pg_catalog.pg_tablespace t ON (t.oid = c.reltablespace)
                 WHERE $whereclause
                 ORDER BY "TABLE_TYPE", "TABLE_CAT", "TABLE_SCHEM", "TABLE_NAME"
                };

            if (defined($type) and length($type) and $type ne '%') {
                my $type_restrict = join ', ' =>
                                      map { /^'/ ? $_ : $dbh->quote($_) }
                                        grep {length}
                                          split(',', $type); ## no critic
                $tbl_sql = qq{SELECT * FROM ($tbl_sql) ti WHERE "TABLE_TYPE" IN ($type_restrict)};
            }
        }
        my $sth = $dbh->prepare($tbl_sql);
        $sth->execute();

        return $sth;
    }

    sub tables {
            my ($dbh, @args) = @_;
            my $attr = $args[4];
            my $sth = $dbh->table_info(@args);
            my $tablelist = $sth->fetchall_arrayref();
            my @tables = map { (! (ref $attr eq 'HASH' and $attr->{pg_noprefix})) ?
                        "$_->[1].$_->[2]" : $_->[2] } @$tablelist;
            return @tables;
    }

    sub table_attributes {

        my ($dbh, $table) = @_;

        my $sth = $dbh->column_info(undef,undef,$table,undef);

        my %convert = (
            COLUMN_NAME   => 'NAME',
            DATA_TYPE     => 'TYPE',
            COLUMN_SIZE   => 'SIZE',
            NULLABLE      => 'NOTNULL',
            REMARKS       => 'REMARKS',
            COLUMN_DEF    => 'DEFAULT',
            pg_constraint => 'CONSTRAINT',
        );

        my $attrs = $sth->fetchall_arrayref(\%convert);

        for my $row (@$attrs) {
            # switch the column names
            for my $name (keys %$row) {
                $row->{ $convert{$name} } = $row->{$name};

                ## Keep some original columns
                delete $row->{$name} unless ($name eq 'REMARKS' or $name eq 'NULLABLE');

            }
            # Moved check outside of loop as it was inverting the NOTNULL value for
            # attribute.
            # NOTNULL inverts the sense of NULLABLE
            $row->{NOTNULL} = ($row->{NOTNULL} ? 0 : 1);

            my @pri_keys = $dbh->primary_key( undef, undef, $table );
            $row->{PRIMARY_KEY} = scalar(grep { /^$row->{NAME}$/i } @pri_keys) ? 1 : 0;
        }

        return $attrs;

    }

    sub _calc_col_size {

        my $mod = shift;
        my $size = shift;


        if ($size > 0) {
            return $size;
        }
        elsif ($mod > 0xffff) {
            my $prec = ($mod & 0xffff) - 4;
            $mod >>= 16;
            my $dig = $mod;
            return "$prec,$dig";
        }
        elsif ($mod >= 4) {
            return $mod - 4;
        }

        return;
    }


    sub type_info_all {

        my $names =
            {
             TYPE_NAME          => 0,
             DATA_TYPE          => 1,
             COLUMN_SIZE        => 2,
             LITERAL_PREFIX     => 3,
             LITERAL_SUFFIX     => 4,
             CREATE_PARAMS      => 5,
             NULLABLE           => 6,
             CASE_SENSITIVE     => 7,
             SEARCHABLE         => 8,
             UNSIGNED_ATTRIBUTE => 9,
             FIXED_PREC_SCALE   => 10,
             AUTO_UNIQUE_VALUE  => 11,
             LOCAL_TYPE_NAME    => 12,
             MINIMUM_SCALE      => 13,
             MAXIMUM_SCALE      => 14,
             SQL_DATA_TYPE      => 15,
             SQL_DATETIME_SUB   => 16,
             NUM_PREC_RADIX     => 17,
             INTERVAL_PRECISION => 18,
            };

        ## This list is derived from dbi_sql.h in DBI, from types.c and types.h, and from the PG docs

        ## Aids to make the list more readable:
        my $GIG = 1073741824;
        my $PS = 'precision/scale';
        my $LEN = 'length';
        my $UN;
        my $ti =
            [
             $names,

['unknown',  SQL_UNKNOWN_TYPE,  0,    $UN,$UN,   $UN,  1,0,0, $UN,0,0, 'UNKNOWN',   $UN,$UN,
             SQL_UNKNOWN_TYPE,                                                             $UN, $UN, $UN ],
['bytea',    SQL_VARBINARY,     $GIG, q{'},q{'}, $UN,  1,0,3, $UN,0,0, 'BYTEA',     $UN,$UN,
             SQL_VARBINARY,                                                                $UN, $UN, $UN ],
['bpchar',   SQL_CHAR,          $GIG, q{'},q{'}, $LEN, 1,1,3, $UN,0,0, 'CHARACTER', $UN,$UN,
             SQL_CHAR,                                                                     $UN, $UN, $UN ],
['numeric',  SQL_DECIMAL,       1000, $UN,$UN,   $PS,  1,0,2, 0,0,0,   'FLOAT',     0,1000,
             SQL_DECIMAL,                                                                  $UN, $UN, $UN ],
['numeric',  SQL_NUMERIC,       1000, $UN,$UN,   $PS,  1,0,2, 0,0,0,   'FLOAT',     0,1000,
             SQL_NUMERIC,                                                                  $UN, $UN, $UN ],
['int4',     SQL_INTEGER,       10,   $UN,$UN,   $UN,  1,0,2, 0,0,0,   'INTEGER',   0,0,
             SQL_INTEGER,                                                                  $UN, $UN, $UN ],
['int2',     SQL_SMALLINT,      5,    $UN,$UN,   $UN,  1,0,2, 0,0,0,   'SMALLINT',  0,0,
             SQL_SMALLINT,                                                                 $UN, $UN, $UN ],
['float4',   SQL_FLOAT,         6,    $UN,$UN,   $PS,  1,0,2, 0,0,0,   'FLOAT',     0,6,
             SQL_FLOAT,                                                                    $UN, $UN, $UN ],
['float8',   SQL_REAL,          15,   $UN,$UN,   $PS,  1,0,2, 0,0,0,   'REAL',      0,15,
             SQL_REAL,                                                                     $UN, $UN, $UN ],
['int8',     SQL_BIGINT,        20,   $UN,$UN,   $UN,  1,0,2, 0,0,0,   'INT8',   0,0,
             SQL_BIGINT,                                                                   $UN, $UN, $UN ],
['date',     SQL_DATE,          10,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'DATE',      0,0,
             SQL_DATE,                                                                     $UN, $UN, $UN ],
['tinterval',SQL_TIME,          18,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TINTERVAL', 0,6,
             SQL_TIME,                                                                     $UN, $UN, $UN ],
['timestamp',SQL_TIMESTAMP,     29,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TIMESTAMP', 0,6,
             SQL_TIMESTAMP,                                                                $UN, $UN, $UN ],
['text',     SQL_LONGVARCHAR,       $GIG, q{'},q{'}, $LEN, 1,1,3, $UN,0,0, 'TEXT',      $UN,$UN,
             SQL_LONGVARCHAR,                                                                  $UN, $UN, $UN ],
['bool',     SQL_BOOLEAN,       1,    q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'BOOLEAN',   $UN,$UN,
             SQL_BOOLEAN,                                                                  $UN, $UN, $UN ],
['array',    SQL_ARRAY,         1,    q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'ARRAY',     $UN,$UN,
             SQL_ARRAY,                                                                    $UN, $UN, $UN ],
['date',     SQL_TYPE_DATE,     10,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'DATE',      0,0,
             SQL_TYPE_DATE,                                                                $UN, $UN, $UN ],
['time',     SQL_TYPE_TIME,     18,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TIME',      0,6,
             SQL_TYPE_TIME,                                                                $UN, $UN, $UN ],
['timestamp',SQL_TYPE_TIMESTAMP,29,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TIMESTAMP', 0,6,
             SQL_TYPE_TIMESTAMP,                                                           $UN, $UN, $UN ],
['timetz',   SQL_TYPE_TIME_WITH_TIMEZONE,
                                29,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TIMETZ',    0,6,
             SQL_TYPE_TIME_WITH_TIMEZONE,                                                  $UN, $UN, $UN ],
['timestamptz',SQL_TYPE_TIMESTAMP_WITH_TIMEZONE,
                                29,   q{'},q{'}, $UN,  1,0,2, $UN,0,0, 'TIMESTAMPTZ',0,6,
             SQL_TYPE_TIMESTAMP_WITH_TIMEZONE,                                             $UN, $UN, $UN ],
        #
        # intentionally omitted: char, all geometric types, internal types
    ];
    return $ti;
    }

    my %get_info_type = (


     116 => ['SQL_ACTIVE_ENVIRONMENTS',             0                         ], ## unlimited
   10021 => ['SQL_ASYNC_MODE',                      2                         ], ## SQL_AM_STATEMENT
     120 => ['SQL_BATCH_ROW_COUNT',                 2                         ], ## SQL_BRC_EXPLICIT
     121 => ['SQL_BATCH_SUPPORT',                   3                         ], ## 12 SELECT_PROC + ROW_COUNT_PROC
       2 => ['SQL_DATA_SOURCE_NAME',                sub { sprintf 'dbi:Pg:%s', shift->{Name} } ],
       3 => ['SQL_DRIVER_HDBC',                     0                         ], ## not applicable
     135 => ['SQL_DRIVER_HDESC',                    0                         ], ## not applicable
       4 => ['SQL_DRIVER_HENV',                     0                         ], ## not applicable
      76 => ['SQL_DRIVER_HLIB',                     0                         ], ## not applicable
       5 => ['SQL_DRIVER_HSTMT',                    0                         ], ## not applicable
       ## Not clear what should go here. Some things suggest 'Pg', others 'Pg.pm'. We'll use DBD::Pg for now
       6 => ['SQL_DRIVER_NAME',                     'DBD::Pg'                 ],
      77 => ['SQL_DRIVER_ODBC_VERSION',             '03.00'                   ],
       7 => ['SQL_DRIVER_VER',                      'DBDVERSION'              ], ## magic word
     144 => ['SQL_DYNAMIC_CURSOR_ATTRIBUTES1',      0                         ], ## we can FETCH, but not via methods
     145 => ['SQL_DYNAMIC_CURSOR_ATTRIBUTES2',      0                         ], ## same as above
      84 => ['SQL_FILE_USAGE',                      0                         ], ## SQL_FILE_NOT_SUPPORTED (this is good)
     146 => ['SQL_FORWARD_ONLY_CURSOR_ATTRIBUTES1', 519                       ], ## not clear what this refers to in DBD context
     147 => ['SQL_FORWARD_ONLY_CURSOR_ATTRIBUTES2', 5209                      ], ## see above
      81 => ['SQL_GETDATA_EXTENSIONS',              15                        ], ## 1+2+4+8
     149 => ['SQL_INFO_SCHEMA_VIEWS',               3932149                   ], ## not: assert, charset, collat, trans
     150 => ['SQL_KEYSET_CURSOR_ATTRIBUTES1',       0                         ], ## applies to us?
     151 => ['SQL_KEYSET_CURSOR_ATTRIBUTES2',       0                         ], ## see above
   10022 => ['SQL_MAX_ASYNC_CONCURRENT_STATEMENTS', 0                         ], ## unlimited, probably
       0 => ['SQL_MAX_DRIVER_CONNECTIONS',          \'SHOW max_connections'   ],
     152 => ['SQL_ODBC_INTERFACE_CONFORMANCE',      1                         ], ## SQL_OIC_LEVEL_1
      10 => ['SQL_ODBC_VER',                        '03.00.0000'              ],
     153 => ['SQL_PARAM_ARRAY_ROW_COUNTS',          2                         ], ## correct?
     154 => ['SQL_PARAM_ARRAY_SELECTS',             3                         ], ## PAS_NO_SELECT
      11 => ['SQL_ROW_UPDATES',                     'N'                       ],
      14 => ['SQL_SEARCH_PATTERN_ESCAPE',           '\\'                      ],
      13 => ['SQL_SERVER_NAME',                     \'SELECT pg_catalog.current_database()' ],
     166 => ['SQL_STANDARD_CLI_CONFORMANCE',        2                         ], ## ??
     167 => ['SQL_STATIC_CURSOR_ATTRIBUTES1',       519                       ], ## ??
     168 => ['SQL_STATIC_CURSOR_ATTRIBUTES2',       5209                      ], ## ??
    9000 => ['9000',                                1                         ], ## can escape placeholders


      16 => ['SQL_DATABASE_NAME',                   \'SELECT pg_catalog.current_database()' ],
      17 => ['SQL_DBMS_NAME',                       'PostgreSQL'              ],
      18 => ['SQL_DBMS_VERSION',                    'ODBCVERSION'             ], ## magic word


      20 => ['SQL_ACCESSIBLE_PROCEDURES',           'Y'                       ], ## is this really true?
      19 => ['SQL_ACCESSIBLE_TABLES',               'Y'                       ], ## is this really true?
      82 => ['SQL_BOOKMARK_PERSISTENCE',            0                         ],
      42 => ['SQL_CATALOG_TERM',                    ''                        ], ## empty = catalogs are not supported
   10004 => ['SQL_COLLATION_SEQ',                   \'SHOW server_encoding'   ],
      22 => ['SQL_CONCAT_NULL_BEHAVIOR',            0                         ], ## SQL_CB_NULL
      23 => ['SQL_CURSOR_COMMIT_BEHAVIOR',          1                         ], ## SQL_CB_CLOSE
      24 => ['SQL_CURSOR_ROLLBACK_BEHAVIOR',        1                         ], ## SQL_CB_CLOSE
   10001 => ['SQL_CURSOR_SENSITIVITY',              1                         ], ## SQL_INSENSITIVE
      25 => ['SQL_DATA_SOURCE_READ_ONLY',           'READONLY'                ], ## magic word
      26 => ['SQL_DEFAULT_TXN_ISOLATION',           'DEFAULTTXN'              ], ## magic word (2 or 8)
   10002 => ['SQL_DESCRIBE_PARAMETER',              'Y'                       ],
      36 => ['SQL_MULT_RESULT_SETS',                'Y'                       ],
      37 => ['SQL_MULTIPLE_ACTIVE_TXN',             'Y'                       ],
     111 => ['SQL_NEED_LONG_DATA_LEN',              'N'                       ],
      85 => ['SQL_NULL_COLLATION',                  0                         ], ## SQL_NC_HIGH
      40 => ['SQL_PROCEDURE_TERM',                  'function'                ], ## for now
      39 => ['SQL_SCHEMA_TERM',                     'schema'                  ],
      44 => ['SQL_SCROLL_OPTIONS',                  8                         ], ## not really for DBD?
      45 => ['SQL_TABLE_TERM',                      'table'                   ],
      46 => ['SQL_TXN_CAPABLE',                     2                         ], ## SQL_TC_ALL
      72 => ['SQL_TXN_ISOLATION_OPTION',            10                        ], ## 2+8
      47 => ['SQL_USER_NAME',                       sub { shift->{CURRENT_USER} } ],


     169  => ['SQL_AGGREGATE_FUNCTIONS',            127                       ], ## all of 'em
     117  => ['SQL_ALTER_DOMAIN',                   31                        ], ## all but deferred
      86  => ['SQL_ALTER_TABLE',                    32639                     ], ## no collate
     114  => ['SQL_CATALOG_LOCATION',               0                         ],
   10003  => ['SQL_CATALOG_NAME',                   'N'                       ],
      41  => ['SQL_CATALOG_NAME_SEPARATOR',         ''                        ],
      92  => ['SQL_CATALOG_USAGE',                  0                         ],
      87  => ['SQL_COLUMN_ALIAS',                   'Y'                       ],
      74  => ['SQL_CORRELATION_NAME',               2                         ], ## SQL_CN_ANY
     127  => ['SQL_CREATE_ASSERTION',               0                         ],
     128  => ['SQL_CREATE_CHARACTER_SET',           0                         ],
     129  => ['SQL_CREATE_COLLATION',               0                         ],
     130  => ['SQL_CREATE_DOMAIN',                  23                        ], ## no collation, no defer
     131  => ['SQL_CREATE_SCHEMA',                  3                         ], ## 1+2 schema + authorize
     132  => ['SQL_CREATE_TABLE',                   13845                     ], ## no collation
     133  => ['SQL_CREATE_TRANSLATION',             0                         ],
     134  => ['SQL_CREATE_VIEW',                    9                         ], ## local + create?
     119  => ['SQL_DATETIME_LITERALS',              65535                     ], ## all?
     170  => ['SQL_DDL_INDEX',                      3                         ], ## create + drop
     136  => ['SQL_DROP_ASSERTION',                 0                         ],
     137  => ['SQL_DROP_CHARACTER_SET',             0                         ],
     138  => ['SQL_DROP_COLLATION',                 0                         ],
     139  => ['SQL_DROP_DOMAIN',                    7                         ],
     140  => ['SQL_DROP_SCHEMA',                    7                         ],
     141  => ['SQL_DROP_TABLE',                     7                         ],
     142  => ['SQL_DROP_TRANSLATION',               0                         ],
     143  => ['SQL_DROP_VIEW',                      7                         ],
      27  => ['SQL_EXPRESSIONS_IN_ORDERBY',         'Y'                       ],
      88  => ['SQL_GROUP_BY',                       2                         ], ## GROUP_BY_CONTAINS_SELECT
      28  => ['SQL_IDENTIFIER_CASE',                2                         ], ## SQL_IC_LOWER
      29  => ['SQL_IDENTIFIER_QUOTE_CHAR',          q{"}                      ],
     148  => ['SQL_INDEX_KEYWORDS',                 0                         ], ## not needed for Pg
     172  => ['SQL_INSERT_STATEMENT',               7                         ], ## 1+2+4 = all
      73  => ['SQL_INTEGRITY',                      'Y'                       ], ## e.g. ON DELETE CASCADE?
      89  => ['SQL_KEYWORDS',                       'KEYWORDS'                ], ## magic word
     113  => ['SQL_LIKE_ESCAPE_CLAUSE',             'Y'                       ],
      75  => ['SQL_NON_NULLABLE_COLUMNS',           1                         ], ## NNC_NOT_NULL
     115  => ['SQL_OJ_CAPABILITIES',                127                       ], ## all
      90  => ['SQL_ORDER_BY_COLUMNS_IN_SELECT',     'N'                       ],
      38  => ['SQL_OUTER_JOINS',                    'Y'                       ],
      21  => ['SQL_PROCEDURES',                     'Y'                       ],
      93  => ['SQL_QUOTED_IDENTIFIER_CASE',         3                         ], ## SQL_IC_SENSITIVE
      91  => ['SQL_SCHEMA_USAGE',                   31                        ], ## all
      94  => ['SQL_SPECIAL_CHARACTERS',             '$'                       ], ## there are actually many more...
     118  => ['SQL_SQL_CONFORMANCE',                4                         ], ## SQL92_INTERMEDIATE ??
      95  => ['SQL_SUBQUERIES',                     31                        ], ## all
      96  => ['SQL_UNION',                          3                         ], ## 1+2 = all


     112  => ['SQL_MAX_BINARY_LITERAL_LEN',         0                         ],
      34  => ['SQL_MAX_CATALOG_NAME_LEN',           0                         ],
     108  => ['SQL_MAX_CHAR_LITERAL_LEN',           0                         ],
      30  => ['SQL_MAX_COLUMN_NAME_LEN',            'NAMEDATALEN'             ], ## magic word
      97  => ['SQL_MAX_COLUMNS_IN_GROUP_BY',        0                         ],
      98  => ['SQL_MAX_COLUMNS_IN_INDEX',           0                         ],
      99  => ['SQL_MAX_COLUMNS_IN_ORDER_BY',        0                         ],
     100  => ['SQL_MAX_COLUMNS_IN_SELECT',          0                         ],
     101  => ['SQL_MAX_COLUMNS_IN_TABLE',           250                       ], ## 250-1600 (depends on column types)
      31  => ['SQL_MAX_CURSOR_NAME_LEN',            'NAMEDATALEN'             ], ## magic word
   10005  => ['SQL_MAX_IDENTIFIER_LEN',             'NAMEDATALEN'             ], ## magic word
     102  => ['SQL_MAX_INDEX_SIZE',                 0                         ],
      33  => ['SQL_MAX_PROCEDURE_NAME_LEN',         'NAMEDATALEN'             ], ## magic word
     104  => ['SQL_MAX_ROW_SIZE',                   0                         ], ## actually 1.6 TB, but too big to represent here
     103  => ['SQL_MAX_ROW_SIZE_INCLUDES_LONG',     'Y'                       ],
      32  => ['SQL_MAX_SCHEMA_NAME_LEN',            'NAMEDATALEN'             ], ## magic word
     105  => ['SQL_MAX_STATEMENT_LEN',              0                         ],
      35  => ['SQL_MAX_TABLE_NAME_LEN',             'NAMEDATALEN'             ], ## magic word
     106  => ['SQL_MAX_TABLES_IN_SELECT',           0                         ],
     107  => ['SQL_MAX_USER_NAME_LEN',              'NAMEDATALEN'             ], ## magic word


      48  => ['SQL_CONVERT_FUNCTIONS',              2                         ], ## CVT_CAST only?
      49  => ['SQL_NUMERIC_FUNCTIONS',              16777215                  ], ## ?? all but some naming clashes: rand(om), trunc(ate), log10=ln, etc.
      50  => ['SQL_STRING_FUNCTIONS',               16280984                  ], ## ??
      51  => ['SQL_SYSTEM_FUNCTIONS',               0                         ], ## ??
     109  => ['SQL_TIMEDATE_ADD_INTERVALS',         0                         ], ## ?? no explicit timestampadd?
     110  => ['SQL_TIMEDATE_DIFF_INTERVALS',        0                         ], ## ??
      52  => ['SQL_TIMEDATE_FUNCTIONS',             1966083                   ],


      53  => ['SQL_CONVERT_BIGINT',                 1830399                    ],
      54  => ['SQL_CONVERT_BINARY',                 1830399                    ],
      55  => ['SQL_CONVERT_BIT',                    0                          ],
      56  => ['SQL_CONVERT_CHAR',                   1830399                    ],
      57  => ['SQL_CONVERT_DATE',                   1830399                    ],
      58  => ['SQL_CONVERT_DECIMAL',                1830399                    ],
      59  => ['SQL_CONVERT_DOUBLE',                 1830399                    ],
      60  => ['SQL_CONVERT_FLOAT',                  1830399                    ],
      61  => ['SQL_CONVERT_INTEGER',                1830399                    ],
     123  => ['SQL_CONVERT_INTERVAL_DAY_TIME',      1830399                    ],
     124  => ['SQL_CONVERT_INTERVAL_YEAR_MONTH',    1830399                    ],
      71  => ['SQL_CONVERT_LONGVARBINARY',          0                          ],
      62  => ['SQL_CONVERT_LONGVARCHAR',            0                          ],
      63  => ['SQL_CONVERT_NUMERIC',                1830399                    ],
      64  => ['SQL_CONVERT_REAL',                   1830399                    ],
      65  => ['SQL_CONVERT_SMALLINT',               1830399                    ],
      66  => ['SQL_CONVERT_TIME',                   1830399                    ],
      67  => ['SQL_CONVERT_TIMESTAMP',              1830399                    ],
      68  => ['SQL_CONVERT_TINYINT',                1830399                    ],
      69  => ['SQL_CONVERT_VARBINARY',              0                          ],
      70  => ['SQL_CONVERT_VARCHAR',                1830399                    ],
     122  => ['SQL_CONVERT_WCHAR',                  0                          ],
     125  => ['SQL_CONVERT_WLONGVARCHAR',           0                          ],
     126  => ['SQL_CONVERT_WVARCHAR',               0                          ],
    ); ## end of %get_info_type
    ## Add keys for names into the hash
    for (keys %get_info_type) {
        $get_info_type{$get_info_type{$_}->[0]} = $get_info_type{$_};
    }

    sub get_info {

        my ($dbh,$type) = @_;

        return undef unless exists $get_info_type{$type};

        my $ans = $get_info_type{$type}->[1];

        if (ref $ans eq 'CODE') {
            $ans = $ans->($dbh);
        }
        elsif (ref $ans eq 'SCALAR') { # SQL
            return $dbh->selectall_arrayref($$ans)->[0][0];
        }
        elsif ($ans eq 'NAMEDATALEN') {
            return $dbh->selectall_arrayref('SHOW max_identifier_length')->[0][0];
        }
        elsif ($ans eq 'ODBCVERSION') {
            my $version = $dbh->{private_dbdpg}{version};
            return '00.00.0000' unless $version =~ /^([0-9][0-9]?)([0-9][0-9])([0-9][0-9])$/;
            return sprintf '%02d.%02d.%.2d00', $1,$2,$3;
        }
        elsif ($ans eq 'DBDVERSION') {
            my $simpleversion = $DBD::Pg::VERSION;
            $simpleversion =~ s/_/./g;
            no if $] >= 5.022, warnings => 'redundant';
            return sprintf '%02d.%02d.%1d%1d%1d%1d', split (/\./, "$simpleversion.0.0.0.0.0.0");
        }
         elsif ($ans eq 'KEYWORDS') {
            ## http://www.postgresql.org/docs/current/static/sql-keywords-appendix.html
            ## Basically, we want ones that are 'reserved' for PostgreSQL but not 'reserved' in SQL:2011
            return join ',' => (qw(ANALYSE ANALYZE ASC CONCURRENTLY DEFERRABLE DESC DO FREEZE ILIKE INITIALLY ISNULL LIMIT NOTNULL PLACING RETURNING VARIADIC VERBOSE));
         }
         elsif ($ans eq 'READONLY') {
             my $SQL = q{SELECT CASE WHEN setting = 'on' THEN 'Y' ELSE 'N' END FROM pg_settings WHERE name = 'transaction_read_only'};
             my $info = $dbh->selectall_arrayref($SQL);
             return $info->[0][0];
         }
         elsif ($ans eq 'DEFAULTTXN') {
             my $SQL = q{SELECT CASE WHEN setting = 'read committed' THEN 2 ELSE 8 END FROM pg_settings WHERE name = 'default_transaction_isolation'};
             my $info = $dbh->selectall_arrayref($SQL);
             return $info->[0][0];
         }

         return $ans;
    } # end of get_info

    sub private_attribute_info {
        return {
                pg_async_status                => undef,
                pg_bool_tf                     => undef,
                pg_int8_as_string              => undef,
                pg_db                          => undef,
                pg_default_port                => undef,
                pg_enable_utf8                 => undef,
                pg_utf8_flag                   => undef,
                pg_errorlevel                  => undef,
                pg_expand_array                => undef,
                pg_host                        => undef,
                pg_INV_READ                    => undef,
                pg_INV_WRITE                   => undef,
                pg_lib_version                 => undef,
                pg_options                     => undef,
                pg_pass                        => undef,
                pg_pid                         => undef,
                pg_placeholder_dollaronly      => undef,
                pg_placeholder_nocolons        => undef,
                pg_placeholder_escaped         => undef,
                pg_port                        => undef,
                pg_prepare_now                 => undef,
                pg_protocol                    => undef,
                pg_server_prepare              => undef,
                pg_server_version              => undef,
                pg_socket                      => undef,
                pg_standard_conforming_strings => undef,
                pg_switch_prepared             => undef,
                pg_user                        => undef,
        };
    }
}


{
    package DBD::Pg::st;

    sub parse_trace_flag {
        return DBD::Pg->parse_trace_flag($_[1]);
    }

    sub bind_param_array {

        ## Binds an array of data to a specific placeholder in a statement
        ## The DBI version is broken, so we implement a near-copy here

        my $sth = shift;
        my ($p_id, $value_array, $attr) = @_;

        ## Bail if the second arg is not undef or an arrayref
        return $sth->set_err(1, "Value for parameter $p_id must be a scalar or an arrayref, not a ".ref($value_array))
            if ref $value_array and ref $value_array ne 'ARRAY';

        ## Bail if the first arg is not a number
        return $sth->set_err(1, q{Can't use named placeholders for non-driver supported bind_param_array})
            unless DBI::looks_like_number($p_id); # because we rely on execute(@ary) here

        ## Store the list of items in the hash (will be undef or an arrayref)
        $sth->{ParamArrays}{$p_id} = $value_array;

        ## If any attribs were passed in, we need to call bind_param
        return $sth->bind_param($p_id, '', $attr) if $attr; ## This is the big change so -w does not complain

        return 1;

    } ## end bind_param_array

     sub private_attribute_info {
        return {
                pg_async                  => undef,
                pg_bound                  => undef,
                pg_current_row            => undef,
                pg_direct                 => undef,
                pg_numbound               => undef,
                pg_cmd_status             => undef,
                pg_oid_status             => undef,
                pg_placeholder_dollaronly => undef,
                pg_placeholder_nocolons   => undef,
                pg_prepare_name           => undef,
                pg_prepare_now            => undef,
                pg_segments               => undef,
                pg_server_prepare         => undef,
                pg_size                   => undef,
                pg_switch_prepared        => undef,
                pg_type                   => undef,
        };
    }

} ## end st section

1;

__END__



